#!/usr/bin/env python
# Copyright (c) 2000-2006 ActiveState Software Inc.
# See the file LICENSE.txt for licensing information.

import sys, os, re, string, stat
import tempfile
from xpcom import components, ServerException, COMException, nsError
import process
import koprocessutils
import logging

log = logging.getLogger('koPHPConfigurator')
#log.setLevel(logging.DEBUG)
#---- components


class koPHPConfigurator:
    _com_interfaces_ = [components.interfaces.koIPHPConfigurator]
    _reg_clsid_ = "2C8CD9F6-DDF4-44e8-89E7-D8B26F42BE26"
    _reg_contractid_ = "@activestate.com/koPHPConfigurator;1"
    _reg_desc_ = "PHP Debugger Configurator"

    def _make_full_version(self,version):
        return '.'.join(version.split('.',2)[:2])

    def _configureDebugger(self, phpInfoEx, inipath, extensiondir=None):
        executable = phpInfoEx.executablePath
        if not executable:
            msg = 'Could not locate any PHP binaries'
            log.warn(msg)
            return msg
        oldini = phpInfoEx.cfg_file_path
        import shutil
        if not os.path.exists(executable):
            msg = 'Executable path %s does not exist' % executable
            log.warn(msg)
            return msg
        if oldini and not os.path.isfile(oldini):
            # we have a directory, check for php.ini inside it
            oldini = os.path.join(oldini, "php.ini")
            if not os.path.isfile(oldini):
                log.warn('INI path %s does not exist' % oldini)
                oldini = None
        
        # get the version so we can figure out
        # which extension to install
        versionText = phpInfoEx.version;
        if not versionText:
            msg = 'Unable to retrieve PHP version'
            log.warn(msg)
            return msg
        version = self._make_full_version(versionText)
        if not version:
            msg = 'Unable to retrieve PHP version [',versionText,'].'
            log.warn(msg)
            return msg
        
        # build the path to the extension in komodo/php
        koDirSvc = components.classes["@activestate.com/koDirs;1"]\
                         .getService(components.interfaces.koIDirs)
        koSupportDir = koDirSvc.supportDir
        # XXX only windows and linux supported here!
        if sys.platform.startswith('win'):
            extensionname = 'php_xdebug.dll'
            extensionlabel = 'zend_extension_ts'
        else:
            extensionname = 'xdebug.so'
            extensionlabel = 'zend_extension'
        
        source_extension_dir = os.path.join(koSupportDir, 'php', 'debugging')
        source_extension = os.path.join(source_extension_dir,version,extensionname)
        if not os.path.exists(source_extension):
            msg = 'Debugger extension for PHP version %s is not available.' % versionText
            log.warn(msg)
            return msg
        
        # copy the old ini to the debug location
        try:
            if not os.path.isdir(inipath):
                os.makedirs(inipath)
        except:
            msg = 'Unable to make directory [%s]' %inipath
            log.warn(msg)
            return msg
        if len(inipath) < 7 or inipath[-7:] != 'php.ini':
            ininame = os.path.join(inipath,'php.ini')
        else:
            ininame = inipath
        contents = ""
        if oldini and oldini != ininame:
            try:
                shutil.copy(oldini,ininame)
            except:
                msg = 'Unable to copy [%s] to [%s]' %(oldini,ininame)
                log.warn(msg)
                return msg
        
            try:
                _stat = os.stat(ininame)
                mode = _stat[stat.ST_MODE]
                writable = mode & stat.S_IWRITE == stat.S_IWRITE
                if not writable:
                    os.chmod(ininame, mode | stat.S_IWUSR)
            except:
                msg = 'Unable to make ini writable [%s]' %inipath
                log.warn(msg)
                return msg
    
            # modify the ini file
            try:
                ini = open(ininame)
                contents = ini.read()
                ini.close()
            except:
                msg = 'Unable to read ini at [%s]' % inipath
                log.warn(msg)
                return msg

            # remove previous configurations of xdebug
            contents=contents.replace('xdebug',';xdebug')
            # disable any other zend extensions, they tend to interfere with debuggers
            # these are not normal php extensions
            contents=contents.replace('zend_extension',';zend_extension')
            # remove old activedebug stuff
            contents=contents.replace('extension=activedebug',';extension=activedebug')
            contents=contents.replace('activedebug',';activedebug')
            contents=contents.replace('cgi.force_redirect',';cgi.force_redirect')
        
        if not extensiondir:
            extensiondir = source_extension_dir
            extensionFilePath = source_extension
        else:
            extensionFilePath = os.path.join(extensiondir, extensionname)

        newcontents = """; xdebug config added by Komodo
%s=%s
xdebug.remote_enable=1
xdebug.remote_handler=dbgp
xdebug.remote_mode=req
xdebug.remote_port=9000
cgi.force_redirect = 0
%s
""" % \
            (extensionlabel, extensionFilePath, contents)
        
        try:
            ini = open(ininame,'w+')
            ini.write(newcontents)
            ini.close()
            if sys.platform[:3] != 'win':
                os.chmod(ininame,0644)
        except:
            msg = 'Unable to write ini at [%s]' % inipath
            log.warn(msg)
            return msg
        
        # copy the debugger extension to the extension dir
        if extensiondir != source_extension_dir:
            try:
                if not os.path.isdir(extensiondir):
                    os.makedirs(extensiondir)
            except:
                msg = 'Unable to create directory [%s]' % extensiondir
                log.warn(msg)
                return msg
            dest = os.path.join(extensiondir,extensionname)
            try:
                shutil.copyfile(source_extension,dest)
                if sys.platform[:3] != 'win':
                    os.chmod(dest,0755)
            except:
                msg = 'Unable to copy [%s] to [%s]'%(source_extension,dest)
                log.warn(msg)
                return msg
        
        phpInfoEx.cfg_file_path = inipath
        # validate that this all worked!
        if phpInfoEx.isDebuggerExtensionLoadable:
            return ''
        msg = 'PHP is configured, but was unable to load the debugger extension at [%s].' % extensionFilePath
        log.warn(msg)
        return msg

    def configureDebugger(self, executable, oldini, inipath, extensiondir):
        # I want a seperate instance of infoex, since we'll be doing some
        # extra special stuff.
        phpInfoEx = components.classes['@activestate.com/koPHPInfoInstance;1'].\
                    createInstance(components.interfaces.koIPHPInfoEx);
        phpInfoEx.executablePath = executable
        phpInfoEx.cfg_file_path = oldini
        return self._configureDebugger(phpInfoEx, inipath, extensiondir)

    def configureDebuggerEx(self, phpInfoEx, inipath, extensiondir):
        if phpInfoEx.isDebuggerExtensionLoadable:
            return ''
        return self._configureDebugger(phpInfoEx, inipath, extensiondir)

    def autoConfigure(self, phpInfoEx):
        if phpInfoEx.isDebuggerExtensionLoadable:
            log.debug("...autoConfigure already configured")
            return ''
        # build the path to the extension in komodo/php
        koDirSvc = components.classes["@activestate.com/koDirs;1"]\
                         .getService(components.interfaces.koIDirs)
        inipath = os.path.join(koDirSvc.hostUserDataDir, "php", phpInfoEx.version)
        log.debug("PHP autoConfigure try configure with inipath %s", inipath)
        err = self._configureDebugger(phpInfoEx, inipath)
        if not err:
            phpInfoEx.cfg_file_path = inipath
        return err
